# Copyright 2020 by Michael Thies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .package_model import OPCRelationship, OPCTargetMode, RELATIONSHIP_TYPE_CORE_PROPERTIES,\
    RELATIONSHIP_TYPE_THUMBNAIL
from .zip_package import ZipPackageReader, ZipPackageWriter
from .core_properties import OPCCoreProperties

DEFAULT_CORE_PROPERTIES_NAME = "/docProps/core.xml"
